; Copyright (c) 2017-2018, Intel Corporation
; 
; Redistribution and use in source and binary forms, with or without
; modification, are permitted provided that the following conditions are met:
; 
;     * Redistributions of source code must retain the above copyright notice,
;       this list of conditions and the following disclaimer.
;     * Redistributions in binary form must reproduce the above copyright
;       notice, this list of conditions and the following disclaimer in the
;       documentation and/or other materials provided with the distribution.
;     * Neither the name of Intel Corporation nor the names of its contributors
;       may be used to endorse or promote products derived from this software
;       without specific prior written permission.
; 
; THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
; AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
; IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
; DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
; FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
; DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
; SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
; CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
; OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
; OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
LIBRARY libIPSec_MB.dll
EXPORTS
    aes_gcm_dec_128_avx_gen2		@1
    aes_gcm_dec_128_avx_gen4		@2
    aes_gcm_dec_128_finalize_avx_gen2	@3
    aes_gcm_dec_128_finalize_avx_gen4	@4
    aes_gcm_dec_128_finalize_sse	@5
    aes_gcm_dec_128_sse			@6
    aes_gcm_dec_128_update_avx_gen2	@7
    aes_gcm_dec_128_update_avx_gen4	@8
    aes_gcm_dec_128_update_sse		@9
    aes_gcm_dec_192_avx_gen2		@10
    aes_gcm_dec_192_avx_gen4		@11
    aes_gcm_dec_192_finalize_avx_gen2	@12
    aes_gcm_dec_192_finalize_avx_gen4	@13
    aes_gcm_dec_192_finalize_sse	@14
    aes_gcm_dec_192_sse			@15
    aes_gcm_dec_192_update_avx_gen2	@16
    aes_gcm_dec_192_update_avx_gen4	@17
    aes_gcm_dec_192_update_sse		@18
    aes_gcm_dec_256_avx_gen2		@19
    aes_gcm_dec_256_avx_gen4		@20
    aes_gcm_dec_256_finalize_avx_gen2	@21
    aes_gcm_dec_256_finalize_avx_gen4	@22
    aes_gcm_dec_256_finalize_sse	@23
    aes_gcm_dec_256_sse			@24
    aes_gcm_dec_256_update_avx_gen2	@25
    aes_gcm_dec_256_update_avx_gen4	@26
    aes_gcm_dec_256_update_sse		@27
    aes_gcm_enc_128_avx_gen2		@28
    aes_gcm_enc_128_avx_gen4		@29
    aes_gcm_enc_128_finalize_avx_gen2	@30
    aes_gcm_enc_128_finalize_avx_gen4	@31
    aes_gcm_enc_128_finalize_sse	@32
    aes_gcm_enc_128_sse			@33
    aes_gcm_enc_128_update_avx_gen2	@34
    aes_gcm_enc_128_update_avx_gen4	@35
    aes_gcm_enc_128_update_sse		@36
    aes_gcm_enc_192_avx_gen2		@37
    aes_gcm_enc_192_avx_gen4		@38
    aes_gcm_enc_192_finalize_avx_gen2	@39
    aes_gcm_enc_192_finalize_avx_gen4	@40
    aes_gcm_enc_192_finalize_sse	@41
    aes_gcm_enc_192_sse			@42
    aes_gcm_enc_192_update_avx_gen2	@43
    aes_gcm_enc_192_update_avx_gen4	@44
    aes_gcm_enc_192_update_sse		@45
    aes_gcm_enc_256_avx_gen2		@46
    aes_gcm_enc_256_avx_gen4		@47
    aes_gcm_enc_256_finalize_avx_gen2	@48
    aes_gcm_enc_256_finalize_avx_gen4	@49
    aes_gcm_enc_256_finalize_sse	@50
    aes_gcm_enc_256_sse			@51
    aes_gcm_enc_256_update_avx_gen2	@52
    aes_gcm_enc_256_update_avx_gen4	@53
    aes_gcm_enc_256_update_sse		@54
    aes_gcm_init_128_avx_gen2		@55
    aes_gcm_init_128_avx_gen4		@56
    aes_gcm_init_128_sse		@57
    aes_gcm_init_192_avx_gen2		@58
    aes_gcm_init_192_avx_gen4		@59
    aes_gcm_init_192_sse		@60
    aes_gcm_init_256_avx_gen2		@61
    aes_gcm_init_256_avx_gen4		@62
    aes_gcm_init_256_sse		@63
    aes_gcm_precomp_128_avx_gen2	@64
    aes_gcm_precomp_128_avx_gen4	@65
    aes_gcm_precomp_128_sse		@66
    aes_gcm_precomp_192_avx_gen2	@67
    aes_gcm_precomp_192_avx_gen4	@68
    aes_gcm_precomp_192_sse		@69
    aes_gcm_precomp_256_avx_gen2	@70
    aes_gcm_precomp_256_avx_gen4	@71
    aes_gcm_precomp_256_sse		@72
    aes_keyexp_128_avx			@73
    aes_keyexp_128_enc_avx		@74
    aes_keyexp_128_enc_sse		@75
    aes_keyexp_128_sse			@76
    aes_keyexp_192_avx			@77
    aes_keyexp_192_enc_avx		@78
    aes_keyexp_192_enc_sse		@79
    aes_keyexp_192_sse			@80
    aes_keyexp_256_avx			@81
    aes_keyexp_256_enc_avx		@82
    aes_keyexp_256_enc_sse		@83
    aes_keyexp_256_sse			@84
    aes_xcbc_expand_key_avx		@85
    aes_xcbc_expand_key_sse		@86
    des_key_schedule			@87
    flush_job_avx			@88
    flush_job_avx2			@89
    flush_job_avx512			@90
    flush_job_sse			@91
    init_mb_mgr_avx			@92
    init_mb_mgr_avx2			@93
    init_mb_mgr_avx512			@94
    init_mb_mgr_sse			@95
    md5_one_block_sse			@96
    queue_size_avx			@97
    queue_size_avx2			@98
    queue_size_avx512			@99
    queue_size_sse			@100
    sha1_one_block_avx			@101
    sha1_one_block_sse			@102
    sha224_one_block_avx		@103
    sha224_one_block_sse		@104
    sha256_one_block_avx		@105
    sha256_one_block_sse		@106
    sha384_one_block_avx		@107
    sha384_one_block_sse		@108
    sha512_one_block_avx		@109
    sha512_one_block_sse		@110
;    sse_sha_ext_usage			@111 ## deprecated
    submit_job_avx			@112
    submit_job_avx2			@113
    submit_job_avx512			@114
    submit_job_nocheck_avx		@115
    submit_job_nocheck_avx2		@116
    submit_job_nocheck_avx512		@117
    submit_job_nocheck_sse		@118
    submit_job_sse			@119
    aes_cmac_subkey_gen_sse             @120
    aes_cmac_subkey_gen_avx             @121
    alloc_mb_mgr                        @122
    free_mb_mgr                         @123
    aes_cfb_128_one_sse                 @124
    aes_cfb_128_one_avx                 @125
    sha1_sse                            @126
    sha1_avx                            @127
    sha1_avx2                           @128
    sha1_avx512                         @129
    sha1_one_block_avx2                 @130
    sha1_one_block_avx512               @131
    md5_one_block_avx			@132
    md5_one_block_avx2			@133
    md5_one_block_avx512		@134
    sha224_one_block_avx2		@135
    sha224_one_block_avx512		@136
    sha256_one_block_avx2		@137
    sha256_one_block_avx512		@138
    sha384_one_block_avx2		@139
    sha384_one_block_avx512		@140
    sha512_one_block_avx2		@141
    sha512_one_block_avx512		@142
    get_next_job_sse                    @143
    get_next_job_avx                    @144
    get_next_job_avx2                   @145
    get_next_job_avx512                 @146
    get_completed_job_sse               @147
    get_completed_job_avx               @148
    get_completed_job_avx2              @149
    get_completed_job_avx512            @150
    aes_keyexp_128_enc_avx2		@151
    aes_keyexp_128_enc_avx512		@152
    aes_keyexp_192_enc_avx2		@153
    aes_keyexp_192_enc_avx512		@154
    aes_keyexp_256_enc_avx2		@155
    aes_keyexp_256_enc_avx512		@156
    aes_xcbc_expand_key_avx2		@157
    aes_xcbc_expand_key_avx512		@158
    aes_cmac_subkey_gen_avx2            @159
    aes_cmac_subkey_gen_avx512          @160
    aes_cfb_128_one_avx2                @161
    aes_cfb_128_one_avx512              @162
    aes_keyexp_128_avx2			@163
    aes_keyexp_128_avx512		@164
    aes_keyexp_192_avx2			@165
    aes_keyexp_192_avx512		@166
    aes_keyexp_256_avx2			@167
    aes_keyexp_256_avx512		@168
    imb_get_version_str                 @169
    imb_get_version                     @170
    init_mb_mgr_sse_no_aesni            @171
    submit_job_sse_no_aesni             @172
    submit_job_nocheck_sse_no_aesni     @173
    flush_job_sse_no_aesni              @174
    queue_size_sse_no_aesni             @175
    get_completed_job_sse_no_aesni      @176
    get_next_job_sse_no_aesni           @177
    aes_cfb_128_one_sse_no_aesni        @178
    aes_gcm_dec_128_finalize_sse_no_aesni       @179
    aes_gcm_dec_128_sse_no_aesni                @180
    aes_gcm_dec_128_update_sse_no_aesni         @181
    aes_gcm_dec_192_finalize_sse_no_aesni       @182
    aes_gcm_dec_192_sse_no_aesni                @183
    aes_gcm_dec_192_update_sse_no_aesni         @184
    aes_gcm_dec_256_finalize_sse_no_aesni       @185
    aes_gcm_dec_256_sse_no_aesni                @186
    aes_gcm_dec_256_update_sse_no_aesni         @187
    aes_gcm_enc_128_finalize_sse_no_aesni       @188
    aes_gcm_enc_128_sse_no_aesni                @189
    aes_gcm_enc_128_update_sse_no_aesni         @190
    aes_gcm_enc_192_finalize_sse_no_aesni       @191
    aes_gcm_enc_192_sse_no_aesni                @192
    aes_gcm_enc_192_update_sse_no_aesni         @193
    aes_gcm_enc_256_finalize_sse_no_aesni       @194
    aes_gcm_enc_256_sse_no_aesni                @195
    aes_gcm_enc_256_update_sse_no_aesni         @196
    aes_gcm_init_128_sse_no_aesni               @197
    aes_gcm_init_192_sse_no_aesni               @198
    aes_gcm_init_256_sse_no_aesni               @199
    aes_gcm_precomp_128_sse_no_aesni            @200
    aes_gcm_precomp_192_sse_no_aesni            @201
    aes_gcm_precomp_256_sse_no_aesni            @202
    aes_keyexp_128_enc_sse_no_aesni             @203
    aes_keyexp_128_sse_no_aesni                 @204
    aes_keyexp_192_enc_sse_no_aesni             @205
    aes_keyexp_192_sse_no_aesni                 @206
    aes_keyexp_256_enc_sse_no_aesni             @207
    aes_keyexp_256_sse_no_aesni                 @208
    aes_xcbc_expand_key_sse_no_aesni            @209
    aes_cmac_subkey_gen_sse_no_aesni            @210
    aes_gcm_pre_128_sse_no_aesni		@211
    aes_gcm_pre_192_sse_no_aesni		@212
    aes_gcm_pre_256_sse_no_aesni		@213
    aes_gcm_pre_128_sse				@214
    aes_gcm_pre_192_sse				@215
    aes_gcm_pre_256_sse				@216
    aes_gcm_pre_128_avx_gen2			@217
    aes_gcm_pre_192_avx_gen2			@218
    aes_gcm_pre_256_avx_gen2			@219
    aes_gcm_pre_128_avx_gen4			@220
    aes_gcm_pre_192_avx_gen4			@221
    aes_gcm_pre_256_avx_gen4			@222
    aes_gcm_pre_128_avx512			@223
    aes_gcm_pre_192_avx512			@224
    aes_gcm_pre_256_avx512			@225
    aes_gcm_pre_128_vaes_avx512			@226
    aes_gcm_pre_192_vaes_avx512			@227
    aes_gcm_pre_256_vaes_avx512			@228
    aes_gcm_dec_128_finalize_avx512       	@229
    aes_gcm_dec_128_avx512                	@230
    aes_gcm_dec_128_update_avx512          	@231
    aes_gcm_dec_192_finalize_avx512      	@232
    aes_gcm_dec_192_avx512               	@233
    aes_gcm_dec_192_update_avx512         	@234
    aes_gcm_dec_256_finalize_avx512       	@235
    aes_gcm_dec_256_avx512                	@236
    aes_gcm_dec_256_update_avx512         	@237
    aes_gcm_enc_128_finalize_avx512       	@238
    aes_gcm_enc_128_avx512                	@239
    aes_gcm_enc_128_update_avx512         	@240
    aes_gcm_enc_192_finalize_avx512       	@241
    aes_gcm_enc_192_avx512                	@242
    aes_gcm_enc_192_update_avx512         	@243
    aes_gcm_enc_256_finalize_avx512       	@244
    aes_gcm_enc_256_avx512                	@245
    aes_gcm_enc_256_update_avx512         	@246
    aes_gcm_init_128_avx512               	@247
    aes_gcm_init_192_avx512               	@248
    aes_gcm_init_256_avx512               	@249
    aes_gcm_precomp_128_avx512            	@250
    aes_gcm_precomp_192_avx512            	@251
    aes_gcm_precomp_256_avx512            	@252
    aes_gcm_dec_128_finalize_vaes_avx512        @253
    aes_gcm_dec_128_vaes_avx512                 @254
    aes_gcm_dec_128_update_vaes_avx512          @255
    aes_gcm_dec_192_finalize_vaes_avx512        @256
    aes_gcm_dec_192_vaes_avx512                 @257
    aes_gcm_dec_192_update_vaes_avx512          @258
    aes_gcm_dec_256_finalize_vaes_avx512        @259
    aes_gcm_dec_256_vaes_avx512                 @260
    aes_gcm_dec_256_update_vaes_avx512          @261
    aes_gcm_enc_128_finalize_vaes_avx512        @262
    aes_gcm_enc_128_vaes_avx512                 @263
    aes_gcm_enc_128_update_vaes_avx512          @264
    aes_gcm_enc_192_finalize_vaes_avx512        @265
    aes_gcm_enc_192_vaes_avx512                 @266
    aes_gcm_enc_192_update_vaes_avx512          @267
    aes_gcm_enc_256_finalize_vaes_avx512        @268
    aes_gcm_enc_256_vaes_avx512                 @269
    aes_gcm_enc_256_update_vaes_avx512          @270
    aes_gcm_init_128_vaes_avx512                @271
    aes_gcm_init_192_vaes_avx512                @272
    aes_gcm_init_256_vaes_avx512                @273
    aes_gcm_precomp_128_vaes_avx512             @274
    aes_gcm_precomp_192_vaes_avx512             @275
    aes_gcm_precomp_256_vaes_avx512             @276
    sha224_sse                                  @277
    sha224_avx                                  @278
    sha224_avx2                                 @279
    sha224_avx512                               @280
    sha256_sse                                  @281
    sha256_avx                                  @282
    sha256_avx2                                 @283
    sha256_avx512                               @284
    sha384_sse                                  @285
    sha384_avx                                  @286
    sha384_avx2                                 @287
    sha384_avx512                               @288
    sha512_sse                                  @289
    sha512_avx                                  @290
    sha512_avx2                                 @291
    sha512_avx512                               @292
    zuc_eea3_iv_gen                             @293
    zuc_eia3_iv_gen                             @294
    zuc_eea3_1_buffer_avx                       @295
    zuc_eea3_1_buffer_sse                       @296
    zuc_eea3_4_buffer_avx                       @297
    zuc_eea3_4_buffer_sse                       @298
    zuc_eea3_n_buffer_avx                       @299
    zuc_eea3_n_buffer_sse                       @300
    zuc_eia3_1_buffer_avx                       @301
    zuc_eia3_1_buffer_sse                       @302
    snow3g_f8_iv_gen                            @303
    snow3g_f9_iv_gen                            @304
    snow3g_f8_1_buffer_avx                      @305
    snow3g_f8_1_buffer_avx2                     @306
    snow3g_f8_1_buffer_bit_avx                  @307
    snow3g_f8_1_buffer_bit_avx2                 @308
    snow3g_f8_1_buffer_bit_sse                  @309
    snow3g_f8_1_buffer_bit_sse_no_aesni         @310
    snow3g_f8_1_buffer_sse                      @311
    snow3g_f8_1_buffer_sse_no_aesni             @312
    snow3g_f8_2_buffer_avx                      @313
    snow3g_f8_2_buffer_avx2                     @314
    snow3g_f8_2_buffer_sse                      @315
    snow3g_f8_2_buffer_sse_no_aesni             @316
    snow3g_f8_4_buffer_avx                      @317
    snow3g_f8_4_buffer_avx2                     @318
    snow3g_f8_4_buffer_sse                      @319
    snow3g_f8_4_buffer_sse_no_aesni             @320
    snow3g_f8_8_buffer_avx                      @321
    snow3g_f8_8_buffer_avx2                     @322
    snow3g_f8_8_buffer_multikey_avx             @323
    snow3g_f8_8_buffer_multikey_avx2            @324
    snow3g_f8_8_buffer_multikey_sse             @325
    snow3g_f8_8_buffer_multikey_sse_no_aesni    @326
    snow3g_f8_8_buffer_sse                      @327
    snow3g_f8_8_buffer_sse_no_aesni             @328
    snow3g_f8_n_buffer_avx                      @329
    snow3g_f8_n_buffer_avx2                     @330
    snow3g_f8_n_buffer_multikey_avx             @331
    snow3g_f8_n_buffer_multikey_avx2            @332
    snow3g_f8_n_buffer_multikey_sse             @333
    snow3g_f8_n_buffer_multikey_sse_no_aesni    @334
    snow3g_f8_n_buffer_sse                      @335
    snow3g_f8_n_buffer_sse_no_aesni             @336
    snow3g_f9_1_buffer_avx                      @337
    snow3g_f9_1_buffer_avx2                     @338
    snow3g_f9_1_buffer_sse                      @339
    snow3g_f9_1_buffer_sse_no_aesni             @340
    snow3g_init_key_sched_avx                   @341
    snow3g_init_key_sched_avx2                  @342
    snow3g_init_key_sched_sse                   @343
    snow3g_init_key_sched_sse_no_aesni          @344
    snow3g_key_sched_size_avx                   @345
    snow3g_key_sched_size_avx2                  @346
    snow3g_key_sched_size_sse                   @347
    snow3g_key_sched_size_sse_no_aesni          @348
    kasumi_f8_iv_gen                            @349
    kasumi_f9_iv_gen                            @350
    kasumi_f8_1_buffer_avx                      @351
    kasumi_f8_1_buffer_bit_avx                  @352
    kasumi_f8_1_buffer_bit_sse                  @353
    kasumi_f8_1_buffer_sse                      @354
    kasumi_f8_2_buffer_avx                      @355
    kasumi_f8_2_buffer_sse                      @356
    kasumi_f8_3_buffer_avx                      @357
    kasumi_f8_3_buffer_sse                      @358
    kasumi_f8_4_buffer_avx                      @359
    kasumi_f8_4_buffer_sse                      @360
    kasumi_f8_n_buffer_avx                      @361
    kasumi_f8_n_buffer_sse                      @362
    kasumi_f9_1_buffer_avx                      @363
    kasumi_f9_1_buffer_sse                      @364
    kasumi_f9_1_buffer_user_avx                 @365
    kasumi_f9_1_buffer_user_sse                 @366
    kasumi_init_f8_key_sched_avx                @367
    kasumi_init_f8_key_sched_sse                @368
    kasumi_init_f9_key_sched_avx                @369
    kasumi_init_f9_key_sched_sse                @370
    kasumi_key_sched_size_avx                   @371
    kasumi_key_sched_size_sse                   @372
